/*
 * Créé à partir de BlackJack.cpp par Shimon Whiteson
 * Auteur : Pierre Gérard pierre.gerard@lipn.univ-paris13.fr
 *
 * Environnement Blackjack semblable à celui décrit dans le Sutton et Barto, section 5.1
 *
 * Observations : int*
 * - int[0] : total du joueur
 * - int[1] : total du donneur
 * - int[3] : le joueur a-t-il un as utilisable ?
 * Actions : int
 * - 0,1 pour déterminer si le joueur arrête ou tire une nouvelle carte
 * Récompense : float
 * - -1 si le jouer perd
 * - 1 si le joueur gagne
 * - 0 en cas d'égalité
 *
 * Tâche épisodique
 */


#include "Blackjack.h"
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

using namespace std;

int playerTotalPrec;
int dealerTotalPrec;

Task_specification env_init()
{
	char* ts = new char[10000];
	strcpy(ts,"0.9:e:3_[i,i,i]_[0,21]_[0,21]_[0,1]:1_[i]_[0,1]");
	return ts;
}

Observation env_start() {
  // initialize reward and observation
  observation = new int[3];
  rewOb.r = 0;
  rewOb.o = observation;
  rewOb.terminal = false;

  // initialize global state variables
  playerTotal = 0;
  dealerTotal = 0;
  playerUsableAce = false;
  dealerUsableAce = false;

  playerTotalPrec = 0;
  dealerTotalPrec = 0;

  // deal player two cards

  playerTotalPrec = playerTotal;
  addCard(playerTotal,playerUsableAce);
  printf("   - ENV : Le joueur  tire un %2d (score : %2d - %2d)\n",playerTotal-playerTotalPrec,playerTotal,dealerTotal); fflush(stdout);
  playerTotalPrec = playerTotal;
  addCard(playerTotal,playerUsableAce);
  printf("   - ENV : Le joueur  tire un %2d (score : %2d - %2d)\n",playerTotal-playerTotalPrec,playerTotal,dealerTotal); fflush(stdout);

  // deal dealer one card (face up)
  dealerTotalPrec = dealerTotal;
  addCard(dealerTotal,dealerUsableAce);
     printf("   - ENV : Le donneur tire un %2d (score : %2d - %2d)\n",dealerTotal-dealerTotalPrec,playerTotal,dealerTotal); fflush(stdout);

  // update and return observation
  setObservation();
  return observation;
}

Reward_observation env_step(Action a)  {  
  // player hits
  if (a == 0) {
    // deal player another card
    playerTotalPrec = playerTotal;
    addCard(playerTotal,playerUsableAce);
    printf("   - ENV : Le joueur  tire un %2d (score : %2d - %2d)\n",playerTotal-playerTotalPrec,playerTotal,dealerTotal); fflush(stdout);

    // check if player is bust
    if (playerTotal > 21) {
      rewOb.r = -1;
      printf("   - ENV : Le joueur PERD\n"); fflush(stdout);
      rewOb.terminal = true;
    }
  }

  // player sticks
  else if (a == 1) {
    printf("   - ENV : Le joueur décide de stopper\n"); fflush(stdout);
    // deal dealer cards until he hits 17 or higher
    while (dealerTotal < 17) {
      dealerTotalPrec = dealerTotal;
      addCard(dealerTotal,dealerUsableAce);
      printf("   - ENV : Le donneur tire un %2d (score : %2d - %2d)\n",dealerTotal-dealerTotalPrec,playerTotal,dealerTotal); fflush(stdout);
    }

    // check to see who won
    if (dealerTotal > 21 || dealerTotal < playerTotal) {
      rewOb.r = 1; // player won
      printf("   - ENV : Le joueur GAGNE\n"); fflush(stdout);
    }
    else if (dealerTotal == playerTotal) {
      rewOb.r = 0; // draw
      printf("   - ENV : La partie est NULLE\n"); fflush(stdout);
    }
    else {
      rewOb.r = -1; // dealer won
      printf("   - ENV : Le joueur PERD\n"); fflush(stdout);
    }
    rewOb.terminal = true;
  }

  // update and return observation and reward
  setObservation();
  return rewOb;
}   

// update global obvservation variable
void setObservation() {
  observation[0] = playerTotal;
  observation[1] = dealerTotal;
  observation[2] = playerUsableAce;
}

// add a new card to a player's total, handle usuable aces
void addCard(int& total, bool& usableAce) {
  // draw a new card
  int card = 1 + rand() % 13;
  
  // if it's a face card, its value is 10
  if (card > 10)
    card = 10;

  // update total
  total += card;

  // if the player went bust, check for a usable ace
  // if it exists, change its value from 11 to 1
  if (total > 21 && usableAce) {
    total -= 10;
    usableAce = false;
  }

  // if the new card is an ace, check if its usable
  if (card == 1 && total < 12) {
    usableAce = true;
    total += 10;
  }
}

void env_cleanup()
{}
