/*
 * Basé sur SarsaAgent.cpp de par Anna Koop
 * Auteur : Pierre Gérard
 * 
 * Agent Q(0)
 * Etats : int* (tableau d'attributs entiers)
 * Actions : int (attribut entier)
 */ 

#include "DynaQPAgent.h"

double gamma = 0.9;

int agent_verbose = false;

void agent_init(Task_specification task_spec)
{
	srand48(999);

	alpha = 0.1;
	epsilon = 0.05;
	nb_simul = 20 ;
	
	task_spec_struct ps;
	parse_task_spec(task_spec, &ps);

	// Trouver le nombre d'états et d'actions, initialiser les diférents tableaux
	obs_dim=(int)ps.obs_dim;
	obs_sizes=new int[obs_dim];
	printf("obs_dim : %d\n",obs_dim);
	n = 1;
	for (int i=0 ; i<obs_dim ; i++) {
		obs_sizes[i]=(int)(ps.obs_maxs[i])-(int)(ps.obs_mins[i])+1;
		n = n*obs_sizes[i];
	}
	m = (int)ps.action_maxs[0]+1;
	
	fprintf(stderr,"\nTask_spec = %s : num_states = %d, num_actions = %d\n",task_spec,n,m);
	nm = n*m;   

	actions = new int[m];//array of possible actions
	for (int i = 0; i < m; i++)
		actions[i] = i;

	// création de la table de valeurs
	Q = new double*[n];
	for (int i = 0; i < n; i++)
		Q[i] = new double[m];

	for (int i = 0; i < n; i++)
		for(int j =0; j < m; j++)
			Q[i][j] = 0;

	// création des tables de transition
	Tnb = new long**[n];
	for (int i = 0; i < n; i++)
		Tnb[i] = new long*[m];

	for (int i = 0; i < n; i++)
		for(int j =0; j < m; j++)
			Tnb[i][j] = new long[n];

	for (int i = 0; i < n; i++)
		for(int j =0; j < m; j++)
			for(int k =0; k < n; k++)
				Tnb[i][j][k] = 0;

	Ttot = new long*[n];
	for (int i = 0; i < n; i++)
		Ttot[i] = new long[m];

	for (int i = 0; i < n; i++)
		for(int j =0; j < m; j++)
			Ttot[i][j] = 0;


	// création des tables des récompenses immédiates
	Rsum = new double*[n];
	for (int i = 0; i < n; i++)
		Rsum[i] = new double[m];

	for (int i = 0; i < n; i++)
		for(int j =0; j < m; j++)
			Rsum[i][j] = 0;

	Rtot = new long*[n];
	for (int i = 0; i < n; i++)
		Rtot[i] = new long[m];

	for (int i = 0; i < n; i++)
		for(int j =0; j < m; j++)
			Rtot[i][j] = 0;
}

double Rew(int sta, int act)
{
	return (double)(Rsum[sta][act]/Rtot[sta][act]);
}

double Trn(int s0, int act, int s1)
{
	return (double)(Tnb[s0][act][s1]/Ttot[s0][act]);
}

int Trn(int s0, int act)
{
	printf("trn1 - ");
	double rand = (int)drand48()*Ttot[s0][act] ;
	double compteur = 0;
	int s1 = -1;	

	for (int i=0 ; (i<n)&&(s1==-1) ; i++)
	{
		compteur+=Tnb[s0][act][s1];
		if (compteur>=rand)
			s1 = i ;
		printf("%d > %f ? ",compteur,rand);
	}
	printf("trn2\n");
	return s1;
}

void print_q(int s)
{
	printf("|");
	for(int i = 0; i < m; i++)
		printf("%2.1f|",Q[s][i]);
}

int observation_index(Observation o)
{
	int idx = o[0];
	for (int i=1 ; i<obs_dim ; i++)
		idx = idx + o[i]*obs_sizes[i];

	return idx;
}

Action agent_start(Observation o)
{
	int s=observation_index(o);

	if (agent_verbose) printf("   - AGT : Observation = %5d (%2d,%2d)\n",s,o[0],o[1]);

	if (agent_verbose) {printf("   - AGT : Q(s,.) = "); print_q(s); printf("\n");}
	a = egreedy(s);
	if (agent_verbose) printf("   - AGT : Action choisie = %d\n",actions[a]);

	A = a;
	S = s;

	simulate_actions();

	if (agent_verbose) printf("\n\n\n");

	return actions[A];
}

Action agent_step(Reward r, Observation o)
{
	static int step_num = 0;
	step_num++;

	//Update one step, return agent's action
	double delta;
	int greedy_a;

	int s=observation_index(o); 

	if (agent_verbose) printf("   - AGT : Observation = %5d (%2d,%2d)\n",s,o[0],o[1]);
	if (agent_verbose) printf("   - AGT : Récompense = %2.0f\n",r);

	Tnb[S][A][s]++;
	Ttot[S][A]++;

	Rsum[S][A] = Rsum[S][A] + r;
	Rtot[S][A]++;

// 	printf("T(%d,%d)=%d",S,A,s);
// 	printf("   T(%d,%d)=%f\n",S,A,r);

	greedy_a = argmax(s);
	delta = r + Q[s][greedy_a] - Q[S][A];

	if (agent_verbose) printf("   - AGT : delta = %f\n",delta);

	Q[S][A] = Q[S][A] + alpha*delta;
	//if (Q[S][A]<0.001) Q[S][A]=0;  // en cas d'oscillations arithmétiques autour de 0

	if (agent_verbose) {printf("   - AGT : Q(s,.) = "); print_q(s); printf("\n");}
	a = egreedy(s);
	if (agent_verbose) printf("   - AGT : Action choisie = %d\n",actions[A]);

	A = a;
	S = s;

	simulate_actions();

	return actions[a];    
}

void agent_end(Reward r)
{
	if (agent_verbose) printf("   - AGT : Récompense = %2.0f\n",r);
	if (agent_verbose) printf("\n\n\n\n\n");

	Rsum[S][A] = Rsum[S][A] + r;
	Rtot[S][A]++;

	double delta;
	int greedy_a;

	delta = r - Q[S][A];

	Q[S][A] = Q[S][A] + alpha*delta;
	//if (Q[S][A]<0.001) Q[S][A]=0; // en cas d'oscillations arithmétiques autour de 0
}

void simulate_actions()
{
	for (int i=0 ; i<nb_simul ; i++)
	{
		int S_rd = (int)(drand48()*n);
		int A_rd = (int)(drand48()*m);

		double r_rd = Rew(S_rd,A_rd);
		int s_rd = Trn(S_rd,A_rd);

		int greedy_a = argmax(s_rd);
		double delta = r_rd + Q[s_rd][greedy_a] - Q[S_rd][A_rd];

		Q[S_rd][A_rd] = Q[S_rd][A_rd] + alpha*delta;
	}
}

void agent_cleanup()
{
	for( int i = 0; i < n; i++)
		delete [] Q[i];
		
	delete [] Q;
	delete [] actions;
}

int egreedy(int s)
{
	//chooses action epsilon-greedily
	double ran = drand48();//static_cast<double>(RAND_MAX);
	if( ran > epsilon )
		return argmax(s);
	else
		return (int)(drand48()*m);
}

int argmax(int s)
{
	//find the index of the maximal action, break ties randomly
	int bestIndex;
	double bestValue, value;
	int numBests = 1;   // OR 1 (one) not SURE
      
	bestIndex = 0;
	bestValue = Q[s][0];	

	for(int i = 1; i < m; i++)
	{
            value = Q[s][i];
            if( value > bestValue )
            {
                bestValue = value;
                bestIndex = i;
                numBests = 1;
            } 
            else if (value == bestValue) 
            {
                numBests = numBests + 1;
                if( (int)(drand48() * numBests) == 0 )
                    bestIndex = i;
            }
	}
	return bestIndex;
}


