/*
 *  MountainCar.cpp
 *
 *  Created by Rich Sutton, modified by Adam White on September 20 2005.
 *  Copyright (c) 2005 UAlberta. All rights reserved.
 *
 *  Description.
 *		This file contains a simulation of the mountain and car dynamic system
 *		EPISODIC
 *
 *  State: 2D Continous
 *  Actions: 1D discrete (0 1 2)
 */
#include "Mountaincar.h"

long step_count;
long episode_count;

long verbose_env_from = 490; 	// épisode à partir duquel l'environement devient verbose

int tempo = 30000;		// temporisation enre deux pas de temps, dans le cas verbose

Task_specification env_init()
{
	o = new double[stateSize];
	episode_count = 0;
	mcar_position = -0.5;
	mcar_velocity = 0.0;

	static char ts[1024];

	sprintf(ts,"1.1:e:2_[f,f]_[%f,%f]_[%f,%f]:1_[i]_[0,2]",
		mcar_min_position,mcar_max_position,
		-mcar_max_velocity,mcar_max_velocity
		);

	return ts ;

// 	Task_specification ret = new char[1000];
// 	char position[20], velocity[20];
// 
// 	sprintf(position,"[%f,%f]",mcar_min_position,mcar_max_position);
// 	sprintf(velocity,"[%f,%f]",-mcar_max_velocity,mcar_max_velocity);
// 
// 	sprintf(ret,"1:e:2_[f,f]_%s_%s:1_[i]_[0,2])",position,velocity);
// 
// 
// 	return ret;
}


Observation env_start()
{
	step_count = 0;
	episode_count ++;

	mcar_position = -0.5;
	mcar_velocity = 0.0;

	o[0] = mcar_position;
	o[1] = mcar_velocity;

	if (episode_count>verbose_env_from) env_print(0); else printf("*"); fflush(stdout);

	return o;
}


Reward_observation env_step(Action a)
{
	step_count++;

	mcar_velocity += (a-1)*0.001 + cos(3*mcar_position)*(-0.0025);
	if (mcar_velocity > mcar_max_velocity) mcar_velocity = mcar_max_velocity;
	if (mcar_velocity < -mcar_max_velocity) mcar_velocity = -mcar_max_velocity;
	mcar_position += mcar_velocity;
	if (mcar_position > mcar_max_position) mcar_position = mcar_max_position;
	if (mcar_position < mcar_min_position) mcar_position = mcar_min_position;
	if (mcar_position==mcar_min_position && mcar_velocity<0) mcar_velocity = 0;

	o[0] = mcar_position;
	o[1] = mcar_velocity;
	ro.o = o;
	ro.r = -1;
	ro.terminal = false;

	if(mcar_goal())
	{
		ro.r = 100;
		ro.terminal = true;
	}
	else if(episode_count > 100000)
	{
		ro.r = -100;
		ro.terminal = true;
	}

	if (episode_count>verbose_env_from) env_print(a); else printf("."); fflush(stdout);

	return ro;
}


void env_cleanup()
{}


bool mcar_goal()				 // Is Car within goal region?
{
	return mcar_position >= mcar_goal_position;
}


void env_print(Action action)
{
	printf("\n");
	printf("Episode = %d, t = %d\n",episode_count,step_count);

	int nb_intervals = 50;

	int pos_car =  (int)(nb_intervals * 
		((mcar_position-mcar_min_position) / (mcar_max_position-mcar_min_position)));

	int pos_goal =  (int)(nb_intervals * 
		((mcar_goal_position-mcar_min_position) / (mcar_max_position-mcar_min_position)));

	char symbol ;
// 	if (mcar_velocity < 0)
// 		symbol = '<';
// 	else if (mcar_velocity > 0)
// 		symbol = '>';
// 	else
// 		symbol = 'o';
	if (action == 0)
		symbol = '<';
	else if (action == 1)
		symbol = '+';
	else
		symbol = '>';
	if (mcar_goal())
		symbol = 'X';

	char* regle = new char[nb_intervals+1];
	for (int i=0 ; i<nb_intervals ; i++)
		regle[i] = '-' ;
	regle[nb_intervals+1] = '\0';

	regle[pos_goal] = 'x';
	regle[pos_car] = symbol ;

	usleep(tempo);
	printf(regle); printf("\n");
}
