/*
 *  carAgent.cpp
 *
 *
 *  Created by Adam White on 24/02/06.
 *  Copyright 2006 __MyCompanyName__. All rights reserved.
 *
 */

#include "TileAgent.h"

#define gamma 1.0

int verbose = false ;

void agent_init(Task_specification ts)
{
	printf(ts); printf("\n");

// 	// pour Mountaincar
	epsilon = 0.1;	
	alpha = 0.1;	
	lambda = 0.9;
	grid_size = 8;
	num_tilings = 16;

// 	//pour Cartpole
// 	epsilon = 0.1;	
// 	alpha = 0.1;	
// 	lambda = 0.8;
// 	grid_size = 4;
// 	num_tilings = 64;

	srand48(111);

	task_spec_struct ps;
	parse_task_spec(ts, &ps);

	obs_dim=(int)ps.obs_dim;
	obs_width=new double[obs_dim];
	obs_min=new double[obs_dim];
	obs_max=new double[obs_dim];
	tile_width=new double[obs_dim];

	num_actions = (int)ps.action_maxs[0]-(int)ps.action_mins[0]+1;
	min_action = (int)ps.action_mins[0];
	max_action = (int)ps.action_maxs[0];
	printf("\nSpécification de la tâche = %s\n",ts);
	printf("- Dimension états = %d, Nb actions = %d (entre %d et %d)\n",obs_dim,num_actions,min_action,max_action);
	
	for (int i=0 ; i<obs_dim ; i++) {
		obs_min[i] = ps.obs_mins[i];
		obs_max[i] = ps.obs_maxs[i];
		obs_width[i]=ps.obs_maxs[i]-ps.obs_mins[i];
		tile_width[i]=obs_width[i]/grid_size;//((num_tilings+2.0)/2.0);
		printf("   - Dimension %d : min=%f, max=%f, width=%f, tileWidth=%f\n",i,ps.obs_mins[i],obs_max[i],obs_width[i],tile_width[i]);
	}
	printf("\n");

	state_vars = new float[obs_dim];
	oldO = new double[obs_dim];
	Q = new double[num_actions];

	F = new int*[num_actions];
	for (int i = 0; i < num_actions; i++)
		F[i] = new int[num_tilings];

	for (int i=0; i<MEMORY_SIZE; i++)
	{
		theta[i]= 0.0;			 // clear memory at start of each run
		e[i] = 0.0;				 // clear all traces
	}
}


Action agent_start(Observation o)
{
	DecayTraces(0.0);			 // clear all traces
	// clear all traces
	loadF(o);					 // compute features
	loadQ();

	if(drand48() <= epsilon)
		oldAction = (int) (drand48() * (num_actions));
	else
		oldAction = argmax(Q);	 // pick argmax action

	return oldAction + min_action;
}


Action agent_step(Reward r, Observation o)
{
	if (verbose) 
	{
		printf("Obs = |");
		for (int i=0 ; i<obs_dim ; i++)
			printf("%f|",o[i]);
		printf("  r=%f\n",r);
	}

	DecayTraces(gamma*lambda);	 // let traces fall

	// optionally clear other traces
	for (int a=0; a<num_actions; a++)
		if (a != oldAction)
			for (int j=0; j<num_tilings; j++)
				ClearTrace(F[a][j]);

	// replace traces
	for (int j=0; j<num_tilings; j++)
		SetTrace(F[oldAction][j],1.0);

	double delta = r - Q[oldAction];

	loadF(o);	// compute features new state
	loadQ();	// compute new state values

	if(drand48() <= epsilon)
		newAction = (int)(drand48()*num_actions);
	else
		{
		//printf("ARGHHHHHHHHHHHHHHHHHHHHHHHHHHH\n");
		newAction = argmax(Q);	 // pick argmax action
		}

	delta += gamma * Q[newAction];

	double temp = (alpha/num_tilings)*delta;
								 // update theta (learn)
	for (int i=0; i<num_nonzero_traces; i++)
	{
		int f = nonzero_traces[i];
		theta[f] += temp * e[f];
	}							 // update theta (learn)

	loadQ(newAction);

	oldAction = newAction;

	if (verbose) printf("Action = %d\n\n",newAction);

	return newAction + min_action;
}


void agent_end(Reward r)
{
	if (verbose) printf("Fin épisode, reward = %f\n\n",r); 

	DecayTraces(gamma*lambda);	 // let traces fall

	// optionally clear other traces
	for (int a=0; a<num_actions; a++)
		if (a != oldAction)
			for (int j=0; j<num_tilings; j++)
				ClearTrace(F[a][j]);
	// replace traces
	for (int j=0; j<num_tilings; j++)
		SetTrace(F[oldAction][j],1.0);

	double delta = r - Q[oldAction];

	double temp = (alpha/num_tilings)*delta;
								 // update theta (learn)
	for (int i=0; i<num_nonzero_traces; i++)
	{
		int f = nonzero_traces[i];
		theta[f] += temp * e[f];
	}
}


void agent_cleanup()
{
}


// Compute all the action values from current F and theta
void loadQ()
{
	for (int a=0; a<num_actions; a++)
	{
		Q[a] = 0;
		for (int j=0; j<num_tilings; j++)
			Q[a] += theta[F[a][j]];
	}
}


// Compute an action value from current F and theta
void loadQ(int a)
{
	Q[a] = 0;
	for (int j=0; j<num_tilings; j++)
		Q[a] += theta[F[a][j]];
}


// Compute feature sets for current car state
void loadF(double* o)
{
	for (int i=0 ; i<obs_dim ; i++)
		state_vars[i] = o[i] / tile_width[i] ;

	for (int a=0; a<num_actions; a++)
		tiles(F[a],num_tilings,MEMORY_SIZE, state_vars,obs_dim,a);
}


// Returns index (action) of largest entry in Q array, breaking ties randomly
int argmax(double* Q)
{
	int best_action = 0;
	double best_value = Q[0];
	int num_ties = 1;			 // actually the number of ties plus 1
	double value;

	for (int a=1; a<num_actions; a++)
	{
		value = Q[a];
		if (value >= best_value)
			if (value > best_value)
		{
			best_value = value;
			best_action = a;
		}
		else
		{
			num_ties++;
			if (0 == (int)(drand48()*num_ties))
			{
				best_value = value;
				best_action = a;
			}
		}
	}
	return best_action;
}


void SetTrace(int f, float new_trace_value)
// Set the trace for feature f to the given value, which must be positive
{								 // trace already exists
	if (e[f] >= minimum_trace) e[f] = new_trace_value;
	else						 // ensure room for new trace
	{
		while (num_nonzero_traces >= MAX_NONZERO_TRACES) IncreaseMinTrace();
		e[f] = new_trace_value;
		nonzero_traces[num_nonzero_traces] = f;
		nonzero_traces_inverse[f] = num_nonzero_traces;
		num_nonzero_traces++;
	}
}


void ClearTrace(int f)
// Clear any trace for feature f
{
	if (!(e[f]==0.0))
		ClearExistentTrace(f,nonzero_traces_inverse[f]);
}


void ClearExistentTrace(int f, int loc)
// Clear the trace for feature f at location loc in the list of nonzero traces
{
	e[f] = 0.0;
	num_nonzero_traces--;
	nonzero_traces[loc] = nonzero_traces[num_nonzero_traces];
	nonzero_traces_inverse[nonzero_traces[loc]] = loc;
}


void DecayTraces(float decay_rate)
// Decays all the (nonzero) traces by decay_rate, removing those below minimum_trace
{
								 // necessary to loop downwards
	for (int loc=num_nonzero_traces-1; loc>=0; loc--)
	{
		int f = nonzero_traces[loc];
		e[f] *= decay_rate;
		if (e[f] < minimum_trace) ClearExistentTrace(f,loc);
	}

}


void IncreaseMinTrace()
// Try to make room for more traces by incrementing minimum_trace by 10%,
// culling any traces that fall below the new minimum
{
	minimum_trace += 0.1 * minimum_trace;
	//cout << "Changing minimum_trace to " << minimum_trace << endl;
								 // necessary to loop downwards
	for (int loc=num_nonzero_traces-1; loc>=0; loc--)
	{
		int f = nonzero_traces[loc];
		if (e[f] < minimum_trace) ClearExistentTrace(f,loc);
	};
}
